@@ -17,7 +17,7 @@
 #  License along with this library; if not, write to the Free Software
 #  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 #
-#  $Id: location.rb,v 0.12 2002/03/04 07:37:12 katayama Exp $
+#  $Id: location.rb,v 0.16 2002/07/02 17:31:33 k Exp $
 #
 
 module Bio
@@ -44,17 +44,19 @@ module Bio
       case location
       when /^[<>]?(\d+)$/				# (A, I) n
 	s = e = $1.to_i
-      when /^<?(\d+)\.\.>?(\d+)$/			# (B, I) n..m
+      when /^[<>]?(\d+)\.\.[<>]?(\d+)$/			# (B, I) n..m
 	s = $1.to_i
 	e = $2.to_i
 	if e - s < 0
-	  raise "[Error] invalid range : #{location}"
+#	  raise "[Error] invalid range : #{location}"
+	  $stderr.puts "[Warning] invalid range : #{location}" if $DEBUG
 	end
-      when /^<?(\d+)\^>?(\d+)$/				# (C, I) n^m
+      when /^[<>]?(\d+)\^[<>]?(\d+)$/			# (C, I) n^m
 	s = $1.to_i
 	e = $2.to_i
 	if e - s != 1
-	  raise "[Error] invalid range : #{location}"
+#	  raise "[Error] invalid range : #{location}"
+	  $stderr.puts "[Warning] invalid range : #{location}" if $DEBUG
 	end
       when /^"?([ATGCatgc]+)"?$/			# (H) literal sequence
 	sequence = $1.downcase
@@ -89,6 +91,10 @@ module Bio
       self					# return Location object
     end
 
+    def range
+      @from..@to
+    end
+
   end
 
 
@@ -110,12 +116,61 @@ module Bio
       end
     end
 
+    def [](n)
+      @locations[n]
+    end
+
     def span
       span_min = @locations.min { |a,b| a.from <=> b.from }
       span_max = @locations.max { |a,b| a.to   <=> b.to   }
       return span_min.from, span_max.to
     end
 
+    def range
+      min, max = span
+      min..max
+    end
+
+    def length
+      len = 0
+      @locations.each do |x|
+	if x.sequence
+	  len += x.sequence.size
+	else
+	  len += (x.to - x.from + 1)
+	end
+      end
+      len
+    end
+    alias :size :length
+
+    def relative(n, type = nil)
+      case type
+      when :location
+	;
+      when :aa
+	if n = abs2rel(n)
+	  (n - 1) / 3 + 1
+	else
+	  nil
+	end
+      else
+	abs2rel(n)
+      end
+    end
+
+    def absolute(n, type = nil)
+      case type
+      when :location
+	;
+      when :aa
+	n = (n - 1) * 3 + 1
+	rel2abs(n)
+      else
+	rel2abs(n)
+      end
+    end
+
 
     private
 
@@ -211,12 +266,60 @@ module Bio
       return ary.flatten
     end
 
+
+    def rel2abs(n) 
+      return nil unless n > 0			# out of range 
+ 
+      cursor = 0 
+      @locations.each do |x|      
+        if x.sequence 
+          len = x.sequence.size 
+        else 
+          len = x.to - x.from + 1 
+        end  
+        if n > cursor + len 
+          cursor += len 
+        else 
+          if x.strand < 0 
+            return x.to - (n - cursor - 1) 
+          else 
+            return x.from + (n - cursor - 1) 
+          end 
+        end                             
+      end 
+      return nil				# out of range 
+    end
+
+    def abs2rel(n)
+      return nil unless n > 0			# out of range
+
+      cursor = 0
+      @locations.each do |x|
+	if x.sequence
+	  len = x.sequence.size
+	else
+	  len = x.to - x.from + 1
+	end
+	if n < x.from or n > x.to then
+	  cursor += len
+	else
+	  if x.strand < 0 then
+	    return x.to - (n - cursor - 1)
+	  else
+	    return n + cursor + 1 - x.from
+	  end
+	end
+      end
+      return nil				# out of range
+    end
+
   end
 
 end
 
 
 if __FILE__ == $0
+  puts "Test new & span methods"
   [
     '754^755',
     'complement(53^54)',
@@ -238,13 +341,41 @@ if __FILE__ == $0
     'one-of(623,627,632)..one-of(628,633,637)',
     'one-of(845,953,963,1078,1104)..1354',
     'join(complement(1..61),complement(AP000007.1:252907..253505))',
+    'complement(join(71606..71829,75327..75446,76039..76203))',
     'order(3..26,complement(964..987))',
     'order(L44135.1:(454.445)..>538,<1..181)',
-  ].each do |x|
-    p x
-    p Bio::Locations.new(x).span
-    p Bio::Locations.new(x)
+    '<200001..<318389',
+  ].each do |pos|
+    p pos
+    p Bio::Locations.new(pos).span
+    p Bio::Locations.new(pos).range
+    p Bio::Locations.new(pos)
   end
+
+  puts "Test rel2abs/abs2rel method"
+  [
+    '6..15',
+    'join(6..10,16..30)',
+    'complement(join(6..10,16..30))',
+    'join(complement(6..10),complement(16..30))',
+    'join(6..10,complement(16..30))',
+  ].each do |pos|
+    loc = Bio::Locations.new(pos)
+    p pos
+#   p loc
+    (1..21).each do |x|
+      print "absolute(#{x}) #=> ", y = loc.absolute(x), "\n"
+      print "relative(#{y}) #=> ", y ? loc.relative(y) : y, "\n"
+      print "absolute(#{x}, :aa) #=> ", y = loc.absolute(x, :aa), "\n"
+      print "relative(#{y}, :aa) #=> ", y ? loc.relative(y, :aa) : y, "\n"
+    end
+  end
+
+  pos = 'join(complement(6..10),complement(16..30))'
+  loc = Bio::Locations.new(pos)
+  print "pos         : "; p pos
+  print "`- loc[1]   : "; p loc[1]
+  print "   `- range : "; p loc[1].range
 end
 
 
@@ -255,7 +386,6 @@ end
 --- Bio::Location.new(location)
 
 --- Bio::Location#from
-
 --- Bio::Location#to
 --- Bio::Location#strand
 --- Bio::Location#sequence
@@ -266,13 +396,56 @@ end
 --- Bio::Location#complement
 --- Bio::Location#replace(sequence)
 
+--- Bio::Location#range
+
+      Returns a Range object of the from..to
+
 = Bio::Locations
 
 --- Bio::Locations.new(position)
 
+      Parse a GenBank style position string and returns a Locations object,
+      which contains a list of Location objects.
+
 --- Bio::Locations#each { |l| ... }
+
+      Iterates on each Location object.
+
+--- Bio::Locations#[](n)
+
+      Returns nth Location object.
+
 --- Bio::Locations#span
 
+      Returns an Array containing overall min and max position [min, max]
+      of this Locations object.
+
+--- Bio::Locations#range
+
+      Similar to span, but returns a Range object min..max
+
+--- Bio::Locations#relative(number, type = nil)
+
+      default
+        absolute position in DNA (na) -> relative position in RNA (na)
+
+      if type == :aa
+        absolute position in DNA (na) -> relative position in Protein (aa)
+
+--- Bio::Locations#absolute(number, type = nil)
+
+      default
+        relative position in RNA (na) -> absolute position in DNA (na)
+
+      if type == :aa
+        relative position in Protein (aa) -> absolute position in DNA (na)
+
+      example
+        loc = Bio::Locations.new('complement(12838..13533)')
+        loc.absolute(10)          #=> 13524 (rel2abs)
+        loc.relative(13524)       #=> 10    (abs2rel)
+        loc.absolute(10, :aa)     #=> 13506 (rel2abs)
+        loc.relative(13506, :aa)  #=> 10    (abs2rel)
 
 == Appendix : GenBank location descriptor classification
 
